/*
*
* 
* 
* 
*/
//package personalizedcommitteeclassifier;

/**
 *
 * @author Md Jamiul Jahid and Jianhua Ruan
 */

// load java packages
import java.util.*;
import java.io.*;
import java.text.*;


// load weka packages
import weka.core.Instances;
import weka.core.Instance;
import weka.classifiers.functions.SMO;
import weka.core.EuclideanDistance;

 
// load jmatio packages, this package is used to read MAT file.
// this package can be downloaded from http://sourceforge.net/projects/jmatio/
import com.jmatio.io.MatFileReader;
import com.jmatio.types.MLArray;
import com.jmatio.types.MLUInt8;


public class PersonalizedCommitteeClassifier {

   
    public final double MAX_VALUE = 1000000.00;
      
    private double [] rwCutoff= {0.001,0.0005,0.0002,0.0001, 0.00005};
    private int neighborsRWInfo[][][];
    
    
    private int [] neighborSizeCross = {1,2,5,10,20,30,40,50,100,150,200,300,400,500,600,700,800,900};
    
    private SMO personalizedSMO [][];
    
    private double pValue[][];
    private int classificationResults [][];
    private double pValueCross[][];
    private int crossClassificationResults [][];
    
    private double weightOfCommittee[];
    private int rowIndexCommittee[];
    private int columnIndexCommittee[];
    
    private double crossPairwiseEuDist [][];
    private int crossPairwiseEuDistIndex [][];
   
    
    public PersonalizedCommitteeClassifier(Instances trainData, Instances testData) {
    
        // initialization of variables    
        neighborsRWInfo = new int[trainData.numInstances()][trainData.numInstances()][rwCutoff.length];
        classificationResults = new int[trainData.numInstances()][rwCutoff.length];
        personalizedSMO = new SMO[trainData.numInstances()][rwCutoff.length];
        pValue = new double[trainData.numInstances()][rwCutoff.length];
        crossPairwiseEuDist = new double[testData.numInstances()][trainData.numInstances()]; 
        crossPairwiseEuDistIndex = new int[testData.numInstances()][trainData.numInstances()]; 
        crossClassificationResults = new int[testData.numInstances()][neighborSizeCross.length];
        pValueCross = new double[testData.numInstances()][neighborSizeCross.length];
     
    }
    
    
    /**
     * sortWithIndex method: insertion sort which also rearrange the index
     */
    public void sortWithIndex(double array[],int index[] )
    {
        //This method implements insertion sort
        for(int i=1;i<array.length;i++)
        {
            
            double item = array[i];
            int ind = index[i];
            int iHole = i;
            
            while((iHole > 0) && (array[iHole-1] > item))
            {
                 // move hole to next smaller index
                array[iHole] = array[iHole-1];
                index[iHole] = index[iHole-1];
                iHole = iHole - 1;
            }
            // put item in the hole
            array[iHole] = item;
            index[iHole] = ind;
            
        }
     }
    
    /**
     * sortWithIndex method: insertion sort which also rearrange the index for first few elements
     */
    public void sortWithIndex(double array[],int indexC[], int indexR[],int size )
    {
        //This method implements insertion sort
        for(int i=1;i<size;i++)
        {
            double item = array[i];
            int indR = indexR[i];
            int indC = indexC[i];
            int iHole = i;
            
            while((iHole > 0) && (array[iHole-1] > item))
            {
                 // move hole to next smaller index
                array[iHole] = array[iHole-1];
                indexR[iHole] = indexR[iHole-1];
                indexC[iHole] = indexC[iHole-1];
                iHole = iHole - 1;
            }
            // put item in the hole
            array[iHole] = item;
            indexR[iHole] = indR;
            indexC[iHole] = indC;
            
        }
     }
            
    /**
     * printMatrixInt method: Write integer type output file
     */
    public void printMatrixInt(int matrix[][],int row, int column, String outputFileName) throws Exception
    {
        
        String outputToFile = new String();
        
        for(int i=0;i<row;i++)
        {
            for(int j=0;j<column;j++)
            {    
                outputToFile = outputToFile+" "+matrix[i][j]+" ";
                System.out.print(" "+matrix[i][j]+" ");
            }  
            outputToFile = outputToFile + "\n";
            System.out.println(" ");
        }
    
        
        //write output to the file
        try {

            FileWriter fstream = new FileWriter(outputFileName);
            BufferedWriter out = new BufferedWriter(fstream);
                
            out.write(outputToFile);
            out.close();
                    
            } catch (FileNotFoundException e) {
            } 
            catch (IOException e) {
                    
            
            }
    
    
    }
    
    /**
     * printMatrixDouble method: Write double type matrix to file 
     */
    public void printMatrixDouble(double matrix[][],int row, int column, String outputFileName) throws Exception
    {
        
        String outputToFile = new String();
        DecimalFormat df = new DecimalFormat("#.###");
        
        for(int i=0;i<row;i++)
        {
            for(int j=0;j<column;j++)
            {    
                double temp = matrix[i][j];
                outputToFile = outputToFile+" "+df.format(temp) +" ";
                System.out.print(" "+df.format(temp)+" ");
            }  
            outputToFile = outputToFile + "\n";
            System.out.println(" ");
        }
    
        
        //write output to the file
        try {

            FileWriter fstream = new FileWriter(outputFileName);
            BufferedWriter out = new BufferedWriter(fstream);
                
            out.write(outputToFile);
            out.close();
                    
            } catch (FileNotFoundException e) {
            } 
            catch (IOException e) {
            
            }
    }
       
    /**
     * calcDistanceCross method: calculate the Euclidean distance of each patient in the testing dataset to the patients in training dataset 
     * @return crossPairwiseEuDist contain the sorted distance and crossPairwiseEuDistIndex contain the index of the sorted patients
     */
    public void calcDistanceCross(Instances data, Instances crossData)
    {
            
        Instances tempData = new Instances(data);
        for(int i=0;i<crossData.numInstances();i++)
        {
            tempData.add(crossData.instance(i));
        }
        
        
        EuclideanDistance ed = new EuclideanDistance(tempData); 
        ed.setDontNormalize(true);  
        
        for(int i=0;i<crossData.numInstances();i++)
        {
            for(int j=0;j<data.numInstances();j++)
            {
                System.out.println(i+" "+j); 
                double dist = ed.distance(crossData.instance(i), data.instance(j));
                 crossPairwiseEuDist[i][j] = dist;
                                    
                 crossPairwiseEuDistIndex[i][j]=j;
              
            }
        }
        //sort the array with their index
        for(int i=0;i<crossData.numInstances();i++)
        {
            sortWithIndex(crossPairwiseEuDist[i],crossPairwiseEuDistIndex[i]);
        }
         
    }
    
    /**
     * calcCommitteeWeight method: sort the all available base classifiers based on the similarity with the testing patients
     */    
    public void calcCommitteeWeight(int instToTest, int sizeDataset)
    {
    
        int correctClassifierCount = 0;
               
        for(int j=0;j<sizeDataset;j++){
            
            int neighbor = crossPairwiseEuDistIndex[instToTest][j];
            double distanceWeight = crossPairwiseEuDist[instToTest][j];
            
            for(int i=0;i<rwCutoff.length;i++){
                
                if(classificationResults[neighbor][i]==1){
                    
                    weightOfCommittee[correctClassifierCount] = distanceWeight;
                    rowIndexCommittee[correctClassifierCount] = neighbor;
                    columnIndexCommittee[correctClassifierCount] = i;
                    correctClassifierCount++;
                }
                
            }
            
        }
        
        sortWithIndex(weightOfCommittee, columnIndexCommittee, rowIndexCommittee, correctClassifierCount);
        
    }
    
    /**
     * testPersonalizedClassifier method: used to predict the label for test patient
     * @param testingData : test dataset
     * @param patientNumber : the index of the patient to test
     * @param neighborSizeIndex : index of how many models used to predict the label of this patient
     * @return store the probability value to pvalueCross matrix
     */ 
    public void testPersonalizedClassifier(Instances testingData, int patientNumber, int neighborSizeIndex) throws Exception
    {
        
        int i, prediction;
        int nM = 0;
        int nNM = 0;
        double resultTemp = 0;
        double result[];
        int numberOfModelsToRun;
        
        // make test dataset
        Instance test = testingData.instance(patientNumber);
          
        numberOfModelsToRun = neighborSizeCross[neighborSizeIndex];
        
        for(i=0;i<numberOfModelsToRun;i++){
            
            // classify this test patient
            result = personalizedSMO[rowIndexCommittee[i]][columnIndexCommittee[i]].distributionForInstance(test);
            
            // store result
            resultTemp = resultTemp+result[1];

            if(result[0]>result[1])
            {    
                nM++;
            }
            else
            {    
                nNM++;
            }
               
        }
       
        if(nM>nNM)
        {    
            prediction = 1;
        }
        else
        {    
            prediction = 0;
        }
       
        // store the average prob-value and classification result
        crossClassificationResults[patientNumber][neighborSizeIndex]= prediction;
        pValueCross[patientNumber][neighborSizeIndex]= resultTemp/(nM+nNM);
        
       
    }
    
    /**
     * makeClassifier method: construct a base classifier based on RW information
     * @param smo : a smo instance of SVM
     * @param trainingData : training dataset
     * @param patientNumber : the representing patient for which the base classifier will be made
     * @param rwCutoffIndex : the cutoff index used to make the base classifier
     * @return base classifier stored in personalizedSMO matrix
     */   
    public void makeClassifier(SMO smo, Instances trainingData, int patientNumber, int rwCutoffIndex) throws Exception
    {
        double result[];
        double prediction;
        
        // make test data
        Instance test = trainingData.instance(patientNumber);
        
        // find the training samples index based on RW information
        int neighborIndex []  = new int[trainingData.numInstances()];
        for(int i=0,j=0;i<trainingData.numInstances();i++){
            if(neighborsRWInfo[i][patientNumber][rwCutoffIndex]==1){
                neighborIndex[j++]=i;
                neighborIndex[j]=-1;
            }
        }
         
        // make training dataset
        Instances training = new Instances(trainingData,neighborIndex[0],1);
        
        int i=1;
        while(neighborIndex[i]!=-1){
            training.add(trainingData.instance(neighborIndex[i]));
            i++;
        }
        
        // set different parameters for the classifier
        String[] options = weka.core.Utils.splitOptions("-M"); //-M for getting distribution
        smo.setOptions(options);
        
        // train the classifier
        smo.buildClassifier(training);
        
        // get the result for the test case
        result = smo.distributionForInstance(test);
        
        // set the p-value to classify this patient        
        pValue[patientNumber][rwCutoffIndex] = result[1];
                
        // get the prediction result to see if it is metastatic or non-metastatic
        if(result[0]>result[1])
        {    
            prediction = 0.0;
        }
        else
        {    
            prediction = 1.0;
        }
        
        // if prediction is correct then include this patient to the classifier 
        // to make it more robust 
        if(prediction==test.classValue())
        {
            classificationResults[patientNumber][rwCutoffIndex]=1;
            
            // new smo to include this patient into it
            SMO smoN = new SMO();
            smoN.setOptions(options);
            training.add(trainingData.instance(patientNumber));
            smoN.buildClassifier(training);
            // save the smo for future patients
            personalizedSMO[patientNumber][rwCutoffIndex] = smoN;
              
        }
        else
        {
            classificationResults[patientNumber][rwCutoffIndex]=0;
            
        }
        
       
        
    }
    
    /**
     * main method
     * @param args containing the training and testing .arff file name
     * @return write the prediction output to files 
     */
    public static void main(String[] args) throws Exception{
                
        //store the testing and training .arff file name 
        String trainF = args[0];
        String testF = args[1];
        
        
        try {
            
            // loading training .arff data file    
            BufferedReader readerTrain = new BufferedReader(new FileReader(trainF));
            Instances trainData = new Instances(readerTrain);
            readerTrain.close();
            trainData.setClassIndex(trainData.numAttributes() - 1); // setting class attribute


            //loading testing .arff data file
            BufferedReader readerTest = new BufferedReader(new FileReader(testF));
            Instances testData = new Instances(readerTest);
            readerTest.close();
            testData.setClassIndex(testData.numAttributes() - 1); // setting class attribute
               

            // make an instance of personalized classifier
            PersonalizedCommitteeClassifier pcc = new PersonalizedCommitteeClassifier(trainData,testData);

              
            
            // loading Random Walk based neighbor information from MAT file and save to a variable 
            for(int i=1;i<=pcc.rwCutoff.length;i++){
                MatFileReader mfr = new MatFileReader("RW_neighbors_"+i+".mat");
                MLArray mlArray = mfr.getMLArray("neighbor");
                       
                for(int j=0;j<trainData.numInstances();j++){
                    for(int k=0;k<trainData.numInstances();k++){
                        pcc.neighborsRWInfo[j][k][i-1]=((MLUInt8)mlArray).get(j,k);
                        }
                    }
                }
                 
               
            //make classification models for training data
            for(int j=0;j<pcc.rwCutoff.length;j++){    
                for(int i=0;i<trainData.numInstances();i++){
                    System.out.println("Making models for paient: "+i+" cutoff: "+pcc.rwCutoff[j]);
                    SMO smo = new SMO();
                    pcc.makeClassifier(smo,trainData,i,j); 
                }
             }


            // calculate similarity based on Euclidean distance and sort the 
            // array for test data
            pcc.calcDistanceCross(trainData,testData);
              

            //test personalized classifier
            for(int i=0;i<testData.numInstances();i++)
            {
                 
                // initialize some variables that is used to rank the models for this patient
                pcc.weightOfCommittee = new double[trainData.numInstances()*pcc.rwCutoff.length];
                pcc.rowIndexCommittee = new int[trainData.numInstances()*pcc.rwCutoff.length];
                pcc.columnIndexCommittee = new int[trainData.numInstances()*pcc.rwCutoff.length];
              
                // rank the models
                pcc.calcCommitteeWeight(i,trainData.numInstances());
                
                // classify patients based on different number of models
                for(int j=0;j<pcc.neighborSizeCross.length;j++){
                
                    pcc.testPersonalizedClassifier(testData,i,j);  
                }
                           
              }
              
                     
            
            // print the result (in this case probability value)
            pcc.printMatrixDouble(pcc.pValueCross,pcc.pValueCross.length, pcc.neighborSizeCross.length,testF.replace("arff","txt"));
            
                    
             
        }
        catch(Exception e)
        {
        
        
        }
        
     
             
       }
}

